#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authserver.py
#
# serveur xmlrpc de gestion de session SSO Eole
#
###########################################################################

import os, shutil, time, traceback
import xmlrpclib
from glob import glob
# imports twisted
from twisted.internet import reactor, defer
# imports sso eole
import config
from page import trace, log
# import de page.py pour pouvoir recharger les templates dans reload
import page
from saml_utils import get_metadata, date_from_string
from saml_crypto import recreate_cert
from saml_utils import available_contexts
from eolesso.dataproxy import LDAPProxy
from eolesso.ticket import AppTicket
from eolesso.util import *
from eolesso.ticketcache import InvalidSession, TicketCache
# import des fonctions de calcul de données utilisateur
import user_infos

if config.USE_SECURID or config.SECURID_PROVIDER:
    from securid_utils import SecuridUserStore

class SSOSessionManager(object):

    def __init__(self):
        self.address = config.AUTH_SERVER_ADDR
        self.init_caches()
        self._parent = None
        self.parent_url = None
        self.load_conf()

    def reload_conf(self):
        reload(user_infos)
        reload(config)
        reload(page)
        self.load_conf()
        return True

    def init_caches(self):
        # cache des "login tickets" (pour empêcher les navigateurs de 'rejouer' l'authentification avec un cache)
        self.login_sessions = TicketCache(180, self.address, 'LT')
        # cache des "granting tickets (ticket de session sso)"
        self.user_sessions = {}
        # cache des attributs utilisateur
        self.data_cache = {}
        # cache des "application" et "proxy" tickets
        self.app_sessions = {}
        # dictionnaire d'association session utilisateur / tickets
        self.user_app_tickets = {}
        # cache pour la gestion des sessions de proxys (expirent en même temps que la session utilisateur)
        self.proxy_granting_sessions = {}
        # cache pour la gestion des sessions de fédération (fournisseur de service)
        self.sp_sessions = {}
        # cache des checkers PAM (Securid)
        self.checkers = {}
        # cache pour la gestion du single logout (logouts en cours)
        # dans le cadre d'une requête de déconnexion globale: stockage de la session sso et du ticket de l'appelant
        # format : {'id_session': ( état_succes_saml,
        #                           [tickets_saml_deco_confirmée],
        #                           (ticket_appelant, id_req_logout, entite_saml_appelante),
        #                           {id_reponse_saml:ticket},
        #                           [services_cas_appelés])
        #          }
        self.pending_logout = {}
        # contextes d'authentification gérés (SAML 2)
        self.allowed_contexts = available_contexts
        if config.USE_SECURID or config.SECURID_PROVIDER:
            self.securid_user_store = SecuridUserStore()
        else:
            self.securid_user_store = None
            del(self.allowed_contexts['URN_TIME_SYNC_TOKEN'])
        if config.SECURID_PROVIDER:
            # cache des requêtes d'authentification OTP distantes en attente
            # Les demandes persistent 5 minutes car la procédure peut prendre
            # un peu de temps (saisie d'un 2ème code)
            self.securid_sessions = TicketCache(300, "", "_")

    def load_conf(self):
        """Charge la configuration du serveur"""
        static_data = {'rne': [config.RNE], 'nom_etab': [config.ETABLISSEMENT]}
        proxy = LDAPProxy(config.LDAP_SERVER, config.LDAP_PORT,
                          config.LDAP_BASE, config.LDAP_LABEL,
                          config.LDAP_READER, config.LDAP_READER_PASSFILE,
                          static_data)
        # lancement du serveur
        if config.PARENT_URL not in ('', None):
            self.parent_url = config.PARENT_URL
            self._parent = xmlrpclib.ServerProxy('%s/xmlrpc' % self.parent_url)
            log.msg("- %s : %s" % (_("Parent server defined"), self.parent_url))
        # proxy d'authentification
        self._data_proxy = proxy
        # attributs calculés
        self.user_infos = user_infos
        log.msg("- %s : %s" % (_("Calculated Attributes defined"), ", ".join(user_infos.dict_infos.keys())))
        self.filters = {}
        self.apps = {'http':{'all':{}},
                     'https':{'all':{}},
                     'all':{'all':{}},
                     'sp_ident':{}}
        self.sp_meta = {}
        self.req_attributes = {}
        self.opt_attributes = {}
        self.attribute_sets = {}
        self.associations = {}
        self.load_filters()
        # chargement des metadata SAML et jeux d'attributs
        self.load_attribute_sets()
        self.init_metadata()

    def check_ticket_issuer(self, ticket_id):
        """vérifie qu'un ticket a été généré par ce serveur
        """
        if ticket_id != '':
            ticket_address = "-".join(ticket_id.split('-')[1:-1])
            if ticket_address == self.address:
                return True
        return False

    def end_session(self, session_id):
        """termine la session d'un utilisateur
        """
        # supression des différents tickets associés
        if session_id in self.user_sessions:
            # recherche des tickets associées à la session
            for ticket in self.user_app_tickets[session_id]:
                # recherche des éventuels PGT associés à cet AppTicket
                if ticket.pgt != None:
                    del self.proxy_granting_sessions[ticket.pgt]
                # recherche des éventuels sessions distantes (saml : idp distant) associées
                if hasattr(ticket,'idp_session_index'):
                    if ticket.idp_session_index in self.sp_sessions:
                        del self.sp_sessions[ticket.idp_session_index]
                # self.user_app_tickets[session_id].remove(ticket)
                del self.app_sessions[ticket.ticket]
            if session_id in self.user_app_tickets:
                del self.user_app_tickets[session_id]
            if session_id in self.data_cache:
                del self.data_cache[session_id]
            if session_id in self.pending_logout:
                del self.pending_logout[session_id]
            del self.user_sessions[session_id]
            log.msg(_("Terminating session %s") % session_id)

    def load_filters(self):
        """chargement des filtres de données et applications
        """
        loaded_apps = []
        loaded_filters = []
        for conf_file in glob('%s/*.ini' % config.FILTER_DIR):
            conf_filename = os.path.splitext(os.path.basename(conf_file))[0]
            cfg = EoleParser()
            cfg.read(conf_file)
            if conf_filename.endswith('_apps') or conf_filename == 'applications':
                # liste des applications associées aux filtres
                loaded_apps.extend(self.load_applications(cfg))
            else:
                # ajout d'un filtre de données
                app_filter = dict( (section, dict(cfg.items(section)))
                                   for section in cfg.sections() )
                self.filters[conf_filename] = app_filter
                loaded_filters.append(conf_filename)
        log.msg("- %s : %s" % (_("Filters loaded"), ", ".join(loaded_filters)))
        log.msg("- %s : %s" % (_("Applications defined"), ", ".join(loaded_apps)))

    def load_applications(self, cfg):
        """Chargement des applications associées aux filtres
        """
        loaded_apps = []
        for app in cfg.sections():
            try:
                app_infos = dict(cfg.items(app))
                # définition des filtres par entité de fournisseur de services (SAML)
                if 'sp_ident' in app_infos:
                    if 'sp_ident' in self.apps['sp_ident']:
                        log.msg(_("Warning, filter for service provider overwritten (already defined): %s") % str(sp_ident))
                    self.apps['sp_ident'][app_infos['sp_ident']] = app_infos['filter']
                    loaded_apps.append(app)
                else:
                    try:
                        port = app_infos['port'] = int(app_infos['port'])
                    except:
                        port = app_infos['port'] = 'all'
                    scheme = app_infos['scheme']
                    if scheme not in ['all', 'both']:
                        if scheme not in self.apps:
                            self.apps[scheme] = {'all':{}}
                        app_stores = [self.apps[scheme]]
                    elif scheme == 'both':
                        app_stores = [self.apps['http'], self.apps['https']]
                    else:
                        # pas de protocole spécifié, on stocke dans tous les protocoles
                        app_stores = self.apps.values()
                    # on trie les applications par port et chemin
                    for store in app_stores:
                        port_store = store.setdefault(port, {})
                        if app_infos['baseurl'] in port_store:
                            stored = False
                            for app_st in port_store[app_infos['baseurl']]:
                                if app_st[1] == app_infos['addr'] and app_st[2] == app_infos['typeaddr']:
                                    log.msg(_("Warning, application overwritten (already defined): %s") % (str(app_st)))
                                    app_st[3] = app_infos['filter']
                                    stored = True
                            if not stored:
                                port_store[app_infos['baseurl']].append([scheme, app_infos['addr'], app_infos['typeaddr'], app_infos['filter']])
                        else:
                            port_store[app_infos['baseurl']] = [[scheme,
                                                                 app_infos['addr'],
                                                                 app_infos['typeaddr'],
                                                                 app_infos['filter']]]
                        loaded_apps.append(app)
            except Exception, e:
                log.msg("! %s %s : %s !" % (_("Error loading application"), app, e))
        return loaded_apps

    def load_attribute_sets(self):
        """chargement des jeux d'attributs pour la fédération d'identité
        """
        for conf_file in glob('%s/*.ini' % config.ATTR_SET_DIR):
            conf_filename = os.path.splitext(os.path.basename(conf_file))[0]
            if not conf_filename.startswith('associations'):
                cfg = EoleParser()
                cfg.read(conf_file)
                # ajout des jeux d'attributs (un seul par fichier)
                # on sépare les attributs en 2 parties:
                # user_attrs : attributs servant à créer le filtre de recherche de l'utilisateur
                # branch_attrs : attributs servant à déterminer un branche de recherche dans l'annuaire
                attr_set = {'user_attrs':{}, 'branch_attrs':{}, 'optional_attrs':{}}
                for section in cfg.sections():
                    if section == 'branch_attrs':
                        attr_set[section] = dict(cfg.items(section))
                    elif section == 'optional':
                        for dist_attr, local_attr in cfg.items(section):
                            attr_set['optional_attrs'][dist_attr] = local_attr
                    else:
                        # tous les autres attributs sont considérés comme requis
                        for dist_attr, local_attr in cfg.items(section):
                            attr_set['user_attrs'][dist_attr] = local_attr
                self.attribute_sets[conf_filename] = attr_set
        # jeu d'attibuts par défaut
        self.associations['default'] = {'attribute_set':'default'}
        for assoc_file in glob('%s/associations*.ini' % config.ATTR_SET_DIR):
            # chargement des associations <entité partenaire><set d'attributs>
            cfg = EoleParser()
            cfg.read(assoc_file)
            for entity_id in cfg.sections():
                # XXX FIXME : permettre d'associer un filtre ici au lieu des filtres avec saml_ident ?
                opts = {'attribute_set':config.ATTR_SET_DIR, 'filter':config.FILTER_DIR}
                for option, src_dir in opts.items():
                    if cfg.has_option(entity_id, option):
                        data_file = os.path.join(src_dir, '%s.ini' % cfg.get(entity_id, option))
                        if option == 'attribute_set' and not os.path.isfile(data_file):
                            log.msg('%s : %s %s' % (assoc_file, data_file, _('not found')))
                        else:
                            assoc = self.associations.get(entity_id, {})
                            assoc[option] = cfg.get(entity_id, option)
                            self.associations[entity_id] = assoc
        if self.associations['default']['attribute_set'] != 'default':
            log.msg(_('default attribute set redefined : %s') % self.associations['default']['attribute_set'])

    def init_metadata(self):
        # réinitialisation des certificats provenant des metadata
        cert_dir = os.path.join(config.METADATA_DIR,'certs')
        if os.path.isdir(cert_dir):
            shutil.rmtree(cert_dir)
        os.makedirs(cert_dir)
        for conf_file in glob('%s/*.xml' % config.METADATA_DIR):
            sp_ident = os.path.splitext(os.path.basename(conf_file))[0]
            self.load_metadata(sp_ident)

    @trace
    def load_metadata(self, sp_ident):
        try:
            sp_meta = get_metadata(sp_ident)
            if 'entityID' in sp_meta:
                sp_ident = sp_meta['entityID']
            if type(sp_ident) == unicode:
                sp_ident = sp_ident.encode(config.encoding)
            f_cert = os.path.join(config.METADATA_DIR,'certs','%s.crt' % sp_ident.replace(os.sep,'_'))
            if not os.path.isfile(f_cert):
                # stockage des certificats sous forme de fichiers pem
                cert_data = None
                if 'SignCert' in sp_meta:
                    cert_data = sp_meta['SignCert']
                elif 'SignCert' in sp_meta.get('SPSSODescriptor',{}):
                    cert_data = sp_meta['SPSSODescriptor']['SignCert']
                elif 'SignCert' in sp_meta.get('IDPSSODescriptor',{}):
                    cert_data = sp_meta['IDPSSODescriptor']['SignCert']
                if cert_data:
                    cert_buffer = recreate_cert(cert_data)
                    # stockage du certificat de signature
                    f = open(f_cert, 'w')
                    f.write(cert_buffer)
                    f.close()
        except Exception, e:
            log.msg("- %s (%s)" % (_("Error fetching SAML metadata for %s : ") % sp_ident, str(e)))
            return {}
        if sp_ident not in self.sp_meta:
            if sp_ident in self.attribute_sets:
                attr_msg = ' (%s)' % _('Custom attribute set')
            else:
                attr_msg = ''
            log.msg("- %s : %s%s" % (_("Partner entity initialized"), sp_ident, attr_msg))
        # mise à jour des métadonnées
        self.sp_meta[sp_ident] = sp_meta
        return sp_meta

    @trace
    def get_metadata(self, sp_ident):
        if sp_ident in self.sp_meta:
            return self.sp_meta[sp_ident]
        else:
            try:
                sp_meta = self.load_metadata(sp_ident)
            except:
                sp_meta = {}
        return sp_meta

    @trace
    def get_attribute_set(self, sp_ident):
        if sp_ident == config.IDP_IDENTITY:
            # cas particulier : utilisation pour retrouver un utilisateur local
            # (si recherche suite à une méthode d'authentification non LDAP)
            return {'user_attrs':{'uid':'uid'}, 'branch_attrs':config.SEARCH_BASE_ATTRS, 'optional_attrs':{}}
        try:
            attr_set = self.associations.get(sp_ident, {}).get('attribute_set', 'default')
            return self.attribute_sets[attr_set]
        except:
            log.msg(_('Unable to determine attribute set for %s') % sp_ident)
            return {'user_attrs':{}, 'branch_attrs':{}, 'optional_attrs':{}}

    @trace
    def init_user_session(self, user_data, username, auth_instant , search_attrs=None, auth_class=available_contexts['URN_PROTECTED_PASSWORD']):
        sessionID = gen_ticket_id('TGC', self.address)
        if sessionID != "":
            # XXX FIXME : utilisation de la clé de fédération comme nom d'utilisateur CAS ?
            if search_attrs is not None:
                self.user_sessions[sessionID] = [search_attrs, auth_instant, auth_class]
            else:
                self.user_sessions[sessionID] = [username, auth_instant, auth_class]
            # stockage des données utilisateur
            for ignored_attr in config.IGNORED_ATTRS:
                if ignored_attr in user_data:
                    del(user_data[ignored_attr])
            self.data_cache[sessionID] = user_data
            # initialisation de la liste des tickets associés
            self.user_app_tickets[sessionID] = []
            reactor.callLater(config.SESSION_TIMER, self.end_session, sessionID)
            log.msg(_("Session initialized (%s)") % username)
        return sessionID

    @trace
    def authenticate_federated_user(self, attrs, idp_ident, auth_instant, auth_class, current_session=None, search_branch=None):
        # recherche du jeu d'attributs à utiliser
        user_data = {}
        attr_set = self.get_attribute_set(idp_ident)
        if search_branch is None:
            search_branch = self._data_proxy.get_search_branch(attrs, attr_set)
        if search_branch is None:
            log.msg(_("Federation - Unable to determine search branch with following attributes : "), attrs)
        else:
            try:
                # si branche non retrouvée, on considère que la fédération a échoué
                search_attrs = {}
                search_base = None
                # on parcourt les attributs reçus
                for idp_attr, value in attrs.items():
                    if idp_attr in attr_set['user_attrs'].keys():
                        # correspond à la clé de fédération
                        if len(value) > 0:
                            search_attrs[attr_set['user_attrs'][idp_attr]] = value[0]
                # tous les attributs doivent correspondre à ceux demandés dans le set d'attributs
                if search_attrs and (len(search_attrs) == len(attr_set['user_attrs'])):
                    defer_fed_data = self._data_proxy.get_user_data(search_attrs, search_branch)
                    return defer_fed_data.addCallbacks(self.callb_federation, self.errb_auth,
                           callbackArgs=[search_attrs, idp_ident, auth_instant, auth_class, current_session])
            except:
                traceback.print_exc()
                user_data = {}
        return defer.succeed(('', user_data))

    @trace
    def callb_federation(self, result, search_attrs, idp_ident, auth_instant, auth_class, current_session):
        bind_success, user_data = result
        sessionID = ''
        if current_session:
            if self.user_sessions[current_session][0] == search_attrs:
                # session déjà existante pour cet utilisateur
                return current_session, user_data
        first_fed_key = search_attrs.keys()[0]
        if first_fed_key in user_data:
            # utilisateur local trouvé: création de la session
            auth_instant = time.mktime(date_from_string(auth_instant).timetuple())
            username = user_data[first_fed_key][0]
            sessionID = self.init_user_session(user_data, username, auth_instant, search_attrs, auth_class)
        else:
            log.msg(_("Federated user from %s not found : %s") % (idp_ident, str(search_attrs)))
        return sessionID, user_data

    def authenticate(self, username, password, search_branch='default'):
        user_data = {}
        sessionID = ''
        if type(username) == unicode: username = username.encode(config.encoding)
        if type(password) == unicode: password = password.encode(config.encoding)

        # on demande en premier au serveur père qu'il vérifie les informations
        if self._parent is not None:
            try:
                sessionID, user_data = self._parent.authenticate(username, password, search_branch)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('authenticate', str(e)))
                pass
        if sessionID != '':
            log.msg(_("Session delivered by parent serveur for %s") % username)
            self.data_cache[sessionID] = user_data
            # session du parent, on crée une session locale 'déportée'
            # on ne gère pas d'expiration sur cette session, la validité sera vérifiée sur le parent
            self.user_sessions[sessionID] = [None,None,None]
            self.user_app_tickets[sessionID] = []
        elif password != '':
            # non authentifié par un parent, on essaye de le faire localement
            defer_auth = self._data_proxy.authenticate(username, password, search_branch)
            return defer_auth.addCallbacks(self.callb_auth, self.errb_auth, callbackArgs=[username])
        return defer.succeed((sessionID, user_data))

    @trace
    def errb_auth(self, failure):
        log.msg(_("! Error accessing Authentication service (%s) !") % self._data_proxy.service_name)
        log.msg(failure.getTraceback())
        return '', {}

    @trace
    def callb_auth(self, res_auth, username):
        """traitement du retour de la méthode d'authentification de dataproxy
        """
        # on stocke le timestamp de la date d'authentification
        auth_instant = time.time()
        # par défaut, on défini la classe d'authentification comme
        # saisie de mot de passe sur un support sécurisé (https)
        # cet attribut sera modifié par le serveur en cas de validation d'un
        # type d'authentification plus élevé (i.e clé OTP).
        auth_class = available_contexts['URN_PROTECTED_PASSWORD']
        authenticated, user_data = res_auth
        if authenticated:
            sessionID = self.init_user_session(user_data, username, auth_instant, auth_class=auth_class)
            return sessionID, user_data
        log.msg(_("! Authentication failure : %s !") % username)
        return '', {}

    def get_login_ticket(self, data=''):
        return self.login_sessions.add_session(data)

    @trace
    def validate_session(self, session_id):
        """vérifie un session_id existant"""
        if self.check_ticket_issuer(session_id) and session_id in self.user_sessions:
            return True
        if self._parent is not None:
            try:
                valid_session = self._parent.validate_session(session_id)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('validate_session', str(e)))
                # XXX FIXME : différencier socket.error ?
                valid_session = False
            if valid_session:
                log.msg(_("User session validated on parent server"))
                # si la session n'est pas présente localement, mais validée par le parent, on l'ajoute (ex : redémarrage du serveur local)
                if not session_id in self.user_sessions:
                    self.user_sessions[session_id] = [None,None,None]
                    self.user_app_tickets[session_id] = []
            return valid_session
        log.msg("! %s : %s !" % (_("Invalid session"), session_id))
        return False

    @trace
    def verify_session_id(self, session_id):
        """vérifie un session_id"""
        if self.check_ticket_issuer(session_id) and session_id in self.user_sessions:
            return True, session_id
        if self._parent is not None:
            try:
                return self._parent.verify_session_id(session_id)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('verify_session', str(e)))
                return False, session_id
        log.msg("! %s : %s !" % (_("Unknown user session"), session_id))
        return False, session_id

    @trace
    def get_user_info(self, session_id, details=False):
        """méthode de convenance pour récupérer l'utilisateur"""
        if session_id in self.user_sessions:
            if self.check_ticket_issuer(session_id):
                if details:
                    return self._get_user_details(session_id, "", {})
                else:
                    return self.user_sessions[session_id][0]
            else:
                try:
                    # session du serveur parent
                    return self._parent.get_user_info(session_id, details)
                except:
                    pass
        raise InvalidSession(session_id)

    @trace
    def get_auth_instant(self, app_ticket):
        """renvoie la date à laquelle l'authentification a eu lieu
        """
        session_id = app_ticket
        ticket = app_ticket
        if app_ticket in self.app_sessions and self.app_sessions[app_ticket] is not None:
            ticket = self.app_sessions[app_ticket]
            session_id = ticket.session_id
            if session_id in self.user_sessions:
                return self.user_sessions[session_id][1]
        else:
            # ticket inconnu, essai de vérification sur un serveur parent
            if self._parent is not None:
                try:
                    return self._parent.get_auth_instant(app_ticket)
                except Exception, e:
                    log.msg(_('Error calling function %s on parent server : %s') % ('verify_app_ticket', str(e)))
                    pass
        raise InvalidSession(session_id)

    @trace
    def get_auth_class(self, app_ticket):
        """retourne la classe d'authentification d'une session"""
        session_id = app_ticket
        ticket = app_ticket
        if app_ticket in self.app_sessions and self.app_sessions[app_ticket] is not None:
            ticket = self.app_sessions[app_ticket]
            session_id = ticket.session_id
            if session_id in self.user_sessions:
                return self.user_sessions[session_id][2]
        if self._parent is not None:
            try:
                return self._parent.get_auth_class(app_ticket)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('get_auth_class', str(e)))
                return None
        log.msg("! %s : %s !" % (_("Unknown user session"), session_id))
        return None

    @trace
    def set_auth_class(self, session_id, auth_class):
        """modifie la classe d'authentification pour une session donnée"""
        if session_id in self.user_sessions:
            self.user_sessions[session_id][2] = auth_class
            return True
        return False

    def logout(self, session_id):
        """vérifie un session_id"""
        if session_id in self.user_sessions:
            self.end_session(session_id)
            if not self.check_ticket_issuer(session_id):
                # session délivrée par le parent ?
                try:
                    return self._parent.logout(session_id)
                except Exception, e:
                    log.msg(_('Error calling function %s on parent server : %s') % ('logout', str(e)))
                    pass
            else:
                return 'ok'
        return ''

    @trace
    def get_app_ticket(self, session_id, appurl, ticket_prefix='ST', from_credentials=False):
        # si cette session utilisateur est valide, on crée un ticket pour l'application
        appSessionID = ''
        if self.validate_session(session_id):
            # génération d'un nouveau ticket d'application associé à la session
            appSessionID = gen_ticket_id(ticket_prefix, self.address)
            ticket = self.init_app_session(appSessionID, session_id, appurl, from_credentials)
            if ticket.filter != '':
                filter_msg = ' (%s : %s)' % (_("attribute filter"), ticket.filter)
            else:
                filter_msg = ''
            log.msg("%s %s%s" % (_("Session authorized for service"), ticket.service_url, filter_msg))
            return appSessionID
        else:
            log.msg(_("Unknown user session"))
        log.msg("! %s %s !" % (_("Failed to create session for service"), appurl))
        return ''

    @trace
    def init_app_session(self, appSessionID, session_id, appurl, from_credentials):
        """génère un ticket d'application et initialise son timeout
        """
        ticket = AppTicket(appSessionID, session_id, appurl, self.address, from_credentials=from_credentials)
        # détection d'un éventuel filtre sur les données utilisateur
        ticket.filter = self._check_filter(ticket.service_url)
        log.msg("%s : %s" % (_("Adding session for service"), ticket.service_url))
        # stockage du ticket
        self.app_sessions[appSessionID] = ticket
        self.user_app_tickets[session_id].append(ticket)
        # mise en place des timeouts de session
        reactor.callLater(config.APP_TIMER, self.invalidate_app_ticket, appSessionID)
        return ticket

    @trace
    def invalidate_app_ticket(self, appSessionID):
        """invalide un ticket d'application pour empêcher son utilisation
        """
        if appSessionID in self.app_sessions:
            self.app_sessions[appSessionID].valid = False

    @trace
    def verify_app_ticket(self, app_ticket, appurl):
        """vérifie le ticket fourni par une application"""
        verified = False
        ticket = app_ticket
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            verified = ticket.verif_ticket(app_ticket, appurl)
        else:
            # ticket inconnu, essai de vérification sur un serveur parent
            if self._parent is not None:
                try:
                    return self._parent.verify_app_ticket(app_ticket, appurl)
                except Exception, e:
                    log.msg(_('Error calling function %s on parent server : %s') % ('verify_app_ticket', str(e)))
                    pass
        if not verified:
            log.msg("! %s %s !" % (_("Session verification failed for service"), get_service_from_url(appurl)))
        return verified, app_ticket

    @trace
    def get_user_details(self, app_ticket, appurl, sections=False, renew=False, keep_valid=False):
        """vérifie un ticket et renvoie les informations sur l'utilisateur
        section : si True, on renvoie aussi une description de l'organisation des données
        """
        log.msg('--- %s %s' % (_("Validating session for"), appurl))
        code = 'INVALID_TICKET'
        detail = "%s : %s" % (_("Unknown ticket"), app_ticket)
        if app_ticket in self.app_sessions:
            log.msg('--- %s' % _("Session OK"))
            ticket = self.app_sessions[app_ticket]
            # si le paramètre renew a été envoyé et que le ticket provient d'une
            # session SSO, on ne le valide pas
            if not ticket.from_credentials and renew:
                detail = _("Ticket %s was not delivered during authentication process") % app_ticket
            elif ticket.verif_ticket(app_ticket, appurl, keep_valid):
                log.msg('--- %s' % _("Appurl verified"))
                return True, self._get_user_details(ticket.session_id, ticket.filter, sections)
            else:
                code = 'INVALID_SERVICE'
                detail = _("Ticket has not been delivered for service %s") % appurl
                log.msg("! %s !" % _("Error reading user data : invalid session"))
        else:
            # session inconnue sur ce serveur, on essaie sur le serveur parent
            if self._parent is not None:
                # récupération des infos sur le serveur parent
                # XXX  FIXME : re-filtrer les infos du serveur parent avec les filtres locaux ??
                try:
                    return self._parent.get_user_details(app_ticket, appurl, sections, renew)
                except Exception, e:
                    log.msg(_('Error calling function %s on parent server : %s') % ('get_user_details', str(e)))
                    pass
        log.msg("! %s %s !" % (_("User data access denied for service"), appurl))
        if sections:
            return False, ({'code':code,'detail':detail},"")
        return False, {'code':code,'detail':detail}

    @trace
    def get_proxy_granting_ticket(self, app_ticket, service, pgturl):
        # si une url de proxy est donnée, on crée un ticket TGC
        # la conformité de l'url est vérifiée lors de la création.
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            if ticket.verif_ticket(app_ticket, service, keep_valid=True):
                ticket.generate_pgt(pgturl)
                self.proxy_granting_sessions[ticket.pgt] = ticket
                return ticket
        return None

    @trace
    def get_proxy_ticket(self, pgt, target_service):
        # récupération de la session
        app_ticket = self.proxy_granting_sessions[pgt]
        appSessionID = gen_ticket_id('PT', self.address)
        # cas spécial : imapproxy ne fonctionne pas avec des pwd de + de 61 caractères
        if target_service.startswith('imap://') or target_service.startswith('imaps://'):
            appSessionID = appSessionID[:60]
        ticket = self.init_app_session(appSessionID, app_ticket.session_id, target_service, app_ticket.from_credentials)
        ticket.parent = app_ticket
        return ticket.ticket

    @trace
    def invalidate_proxy(self, app_ticket):
        """Supprime les paramètres de proxy d'un app_ticket si les vérifications sur l'url de callback ont échoué
        """
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            if ticket.pgt is not None:
                del self.proxy_granting_sessions[ticket.pgt]
            ticket.reset_pgt()

    @trace
    def _check_filter(self, appurl):
        """vérifie si un filtre est disponible pour l'url"""
        url = urlparse.urlparse(appurl)
        if url.scheme in self.apps:
            store = self.apps[url.scheme]
        else:
            store = self.apps['http']
        # recherche d'une application avec le même port/chemin
        if url.port in store:
            apps = store[url.port]
        else:
            # si pas de port correspondant, on regarde dans les applis sans port spécifique
            apps = store['all']
        # gestion du path, il faut que le path commence par ce qu'on veut
        # on trie les chemins connus pour commencer par le plus spécifique
        app_paths = apps.keys()
        app_paths.sort(reverse=True)
        url_path = url.path
        if not url_path.startswith('/'): url_path = '/' + url_path
        for baseurl in app_paths:
            if not baseurl.startswith('/'): check_baseurl = '/' + baseurl
            check_baseurl = baseurl
            if url_path.startswith(check_baseurl):
                for app in apps[baseurl]:
                    scheme, addr, typeaddr, id_filter = app
                    if typeaddr == 'ip':
                        addr_ok = check_hostname_by_ip(url, addr)
                    else:
                        addr_ok = check_hostname_by_domain(url, addr)
                    if addr_ok:
                        # on a une appli correspondant à l'url, on renvoie le filtre associé
                        log.msg(_("Applied filter %s for url %s") % (id_filter, get_service_from_url(appurl)))
                        return id_filter
                        break
        return "default"

    @trace
    def _filter_data(self, infos, id_filter, sections):
        """filtrage en fonction de l'application si des filtres sont définis"""
        data = {}
        if id_filter in self.filters:
            for section, attrs in self.filters[id_filter].items():
                for libelle_cas, nom_val in attrs.items():
                    data[nom_val] = infos.get(nom_val, '')
            filter_data = self.filters[id_filter]
        else:
            # FIXME : définir les données à renvoyer par défaut !
            data.update(infos)
            filter_data = ""
        if sections:
            return data, filter_data
        else:
            return data

    @trace
    def _get_user_details(self, session_id, id_filter, sections):
        """Renvoie les informations connues de l'utilisateur lié à <session_id>"""
        # on récupère l'id de l'utilisateur
        # user_id = self.user_sessions[session_id]
        if self.check_ticket_issuer(session_id):
            infos = self.data_cache[session_id]
        else:
            # session distante
            try:
                infos = self._parent.get_user_info(session_id, True)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('get_user_info', str(e)))
                infos = {}
        # ajout des informations statiques
        infos.update({'rne':[config.RNE],'nom_etab':[config.ETABLISSEMENT]})
        # lancement des fonctions de calcul de valeurs supplémentaires
        for varname, calc_func in self.user_infos.dict_infos.items():
            try:
                infos[varname] = calc_func(infos)
            except Exception, e:
                log.msg("""%s '%s' :  %s""" % (_("Error computing data for attribute"), varname, str(e)))
        infos = self._filter_data(infos, id_filter, sections)
        log.msg('--- %s' % _("User data sent"))
        return infos
